
#include "BoD.h"
#include <time.h>

#define MAX_STATEMENT_LENGTH  100
#define ONE_WEEK (7*24*60*60)    /* in seconds */

int exit_edit_mode = FALSE ;

static int  valid_account = FALSE ;

static int  num_cheques   = 10 ;

static char acct_fname  [ 1024 ] = " " ;
static char acct_name   [ 1024 ] = "No Account Selected" ;
static char addrs_line1 [ 1024 ] = " " ;
static char addrs_line2 [ 1024 ] = " " ;
static time_t allowance_time = 0L ;
static int  allowance     = 0 ;
static int  balance       = 0 ;
static int  next_cheque   = 0 ;
static int  interest_rate = 0 ;

static char acct_statement [ MAX_STATEMENT_LENGTH ][ 1024 ] ;
static int  acct_statement_pointer = 0 ;

void emptyStatement ()
{
  for ( int i = 0 ; i < MAX_STATEMENT_LENGTH ; i++ )
    acct_statement [ i ][ 0 ] = '\0' ;

  acct_statement_pointer = 0 ;
}

void addToStatement ( char *entry )
{
  /*
    If statement got to be too long, throw away
    oldest entry.
  */

  if ( acct_statement_pointer >= MAX_STATEMENT_LENGTH )
  {
    for ( int i = 1 ; i < acct_statement_pointer ; i++ )
      strcpy ( acct_statement[i-1], acct_statement[i] ) ;

    acct_statement_pointer -- ;
  }

  strcpy ( acct_statement [ acct_statement_pointer ], entry ) ;
  acct_statement_pointer++ ;
}


void addToStatement ( time_t t, char *comment, int amount )
{
  char s [ 1024 ] ;
  int a1 = amount  ;
  int b1 = balance ;

  strcpy ( s, ctime(&t) ) ;
  s[10]='\0';         /* Chop off trailing '\n' and year */

  sprintf ( &(s[strlen(s)]), ": %s ", comment ) ;

  while ( strlen ( s ) < 55 )
    strcat ( s, "_" ) ;

  if ( a1 < 0 ) a1 = -a1 ;
  if ( b1 < 0 ) b1 = -b1 ;

  sprintf ( &(s[strlen(s)]), "%c%2d.%02d  Bal: %c%3d.%02d%s\n",
                (amount < 0) ? '-' : '+',
                a1/100, a1-(a1/100)*100,
                (balance < 0) ? '-' : ' ',
                b1/100, b1-(b1/100)*100,
                (balance < 0) ? "!!!" : "" ) ;

  addToStatement ( s ) ;
}



void printPostscriptStatement ()
{
  char *printer = getenv ( "BOD_PRINT_COMMAND" ) ;

  if ( printer == NULL || printer[0] == '\0' )
    printer = "lpr" ;

  FILE *fd = popen ( printer, "w" ) ;

  if ( fd == NULL )
  {
    fprintf ( stderr, "Failed to start printer command '%s'.\n",
                                        printer ) ;
    exit ( 1 ) ;
  }

  int y = 720 ;

  fprintf ( fd, "%!PS-Adobe-2.0 EPSF-2.0\n" ) ;
  fprintf ( fd, "%%Title: BoDstatement.ps\n" ) ;
  fprintf ( fd, "%%Creator: BoD (Bank of Daddy) by Steve Baker\n" ) ;
  fprintf ( fd, "%%Pages: 1\n" ) ;
  fprintf ( fd, "%%DocumentFonts:\n" ) ;
  fprintf ( fd, "%%EndComments\n" ) ;
  fprintf ( fd, "%%EndProlog\n" ) ;
  fprintf ( fd, "%%Page: 1 1\n" ) ;

  fprintf ( fd, "72 %d moveto\n", y ) ;  y -= 18 ;
  fprintf ( fd, "/Times-Bold findfont 15 scalefont setfont\n" ) ;
  fprintf ( fd, "(Bank of Daddy - Statement of Account %s) show\n",
                     acct_name ) ;

  fprintf ( fd, "/Times-Roman findfont 12 scalefont setfont\n" ) ;

  fprintf ( fd, "72 %d moveto\n", y ) ;  y -= 14 ;
  fprintf ( fd, "(Your Allowance is $%d.%02d) show\n", allowance/100,
              allowance - (allowance/100)*100 ) ;

  fprintf ( fd, "72 %d moveto\n", y ) ;  y -= 14 ;
  fprintf ( fd, "(Your Interest Rate is %d%%) show\n", interest_rate ) ;

  fprintf ( fd, "/Courier findfont 10 scalefont setfont\n" ) ;
  fprintf ( fd, "72 %d moveto\n", y ) ;  y -= 14 ;

  for ( int i = 0 ; i < MAX_STATEMENT_LENGTH ; i++ )
  {
    fprintf ( fd, "72 %d moveto\n", y ) ;  y -= 12 ;
    fprintf ( fd, "(%s) show\n", acct_statement [ i ] ) ;
  }

  fprintf ( fd, "72 %d moveto\n", y ) ;  y -= 14 ;
  if ( balance >= 0 )
    fprintf ( fd, "(Current Balance = $%d.%02d) show\n", balance/100,
                    balance - (balance/100)*100 ) ;
  else
    fprintf ( fd, "(**OVERDRAWN BY $%d.%02d) show\n", (-balance)/100,
                    (-balance) - ((-balance)/100)*100 ) ;


  fprintf ( fd, "showpage\n" ) ;
  pclose ( fd ) ;
}

void printStatementRaw ( FILE *fd )
{
  for ( int i = 0 ; i < MAX_STATEMENT_LENGTH ; i++ )
    fprintf ( fd, "%s", acct_statement [ i ] ) ;
}

void updateGUI ()
{
  char bal [ 1024 ] ;

  if ( balance > 0 )
    sprintf(bal,"$%d.%02d", balance/100, balance-(balance/100)*100 ) ;
  else
    sprintf(bal,"$-%d.%02d", (-balance)/100, (-balance)-((-balance)/100)*100) ;

  curr_gui -> Balance     -> value ( bal ) ;
  curr_gui -> AccountName -> value ( acct_name ) ;
}

void fail ( char *s, int line )
{
  char msg [ 1024 ] ;

  sprintf ( msg, "Syntax error '%s' in '%s' line %d", s, acct_fname, line ) ;

  fl_alert ( msg ) ;

  valid_account = FALSE ;
  strcpy ( acct_name, " " ) ;
  balance = 0 ;
  updateGUI () ;
}



int check_account_name ()
{
  if ( ! valid_account )
    fl_alert ( "You must enter a valid account name first." ) ;

  return valid_account ;
}

void setChequeCount ( int c )
{
  if ( c <  1 ) c = 1 ;
  if ( c > 50 ) c = 50 ;

  num_cheques = c ;
}


void printChequeBook ()
{
  if ( ! check_account_name () ) return ;

  if ( ! fl_ask ( "Print ChequeBook?" ) )
    return ;

  printCheque ( num_cheques, next_cheque,
                acct_name, addrs_line1, addrs_line2 ) ;
  next_cheque += num_cheques ;
  refreshAccount () ;
  fl_alert ( "Printing..." ) ;
}


void editAccount ()
{
  if ( ! check_account_name () ) return ;

  EditorInterface *ed_gui = new EditorInterface ( 0, 0, 100, 100 ) ;
  ed_gui   -> end  () ;
  ed_gui   -> interest_rate -> value ( interest_rate ) ;
  ed_gui   -> allowance     -> value ( (float)allowance / 100.0f ) ;
  ed_gui   -> next_cheque_number -> value ( next_cheque ) ;
  ed_gui   -> account_name  -> value ( acct_name   ) ;
  ed_gui   -> line1_name    -> value ( addrs_line1 ) ;
  ed_gui   -> line2_name    -> value ( addrs_line2 ) ;
  ed_gui   -> mainwindow -> set_modal() ;
  ed_gui   -> show () ;
  exit_edit_mode = FALSE ;

  while ( ! exit_edit_mode )
  {
    if ( ! Fl::check () )
    {
      refreshAccount () ;
      delete ed_gui ;
      return ;
    }

    ed_gui->redraw () ;
  }

  interest_rate = (int) ed_gui -> interest_rate      -> value () ;
  allowance     = (int) ( ed_gui -> allowance        -> value () * 100.0f ) ;
  next_cheque   = (int) ed_gui -> next_cheque_number -> value () ;

  strcpy ( acct_name, ed_gui -> account_name  -> value () ) ;
  strcpy ( addrs_line1, ed_gui -> line1_name  -> value () ) ;
  strcpy ( addrs_line2, ed_gui -> line2_name  -> value () ) ;

  refreshAccount () ;
  delete ed_gui->mainwindow ;
  delete ed_gui ;
}


void refreshAccount ()
{
  if ( ! check_account_name () ) return ;

  FILE *fd = fopen ( acct_fname, "w" ) ;

  if ( fd == NULL )
  {
    valid_account = FALSE ;
    strcpy ( acct_name, " " ) ;
    balance = 0 ;
    updateGUI () ;
    return ;
  }

  int  dollars = balance   / 100 ;
  int  cents   = balance   - (  dollars * 100 ) ;
  int adollars = allowance / 100 ;
  int acents   = allowance - ( adollars * 100 ) ;

  fprintf ( fd, "Name=[%s]\n", acct_name ) ;
  fprintf ( fd, "NextCheque=%d\n", next_cheque ) ;
  fprintf ( fd, "Addrs1=[%s]\n", addrs_line1 ) ;
  fprintf ( fd, "Addrs2=[%s]\n", addrs_line2 ) ;
  fprintf ( fd, "NextAllowanceDue=%ld\n", allowance_time ) ;
  fprintf ( fd, "WeeklyInterestRate=%d%%\n" , interest_rate ) ;
  fprintf ( fd, "Allowance=$%d.%d\n", adollars, acents ) ;
  fprintf ( fd, "Balance=$%d.%d\n" , dollars, cents ) ;

  printStatementRaw ( fd ) ;
  fclose ( fd ) ;

  updateGUI () ;
}


void exitBoD ()
{
  if ( valid_account )
    refreshAccount () ;

  exit ( 0 ) ;
}




void selectAccount ()
{
  char *home = getenv ( "BOD_HOME" ) ;
  if ( home == NULL || home[0] == '\0' )
    home = "." ;

  char *tmp = fl_file_chooser ( "Name of Account:",
                                "*.bod",
                                home ) ;

  if ( tmp == NULL || tmp[0] == '\0' )
  {
    valid_account = FALSE ;
    strcpy ( acct_name, " " ) ;
    balance = 0 ;
    updateGUI () ;
    return ;
  }

  strcpy ( acct_fname, tmp ) ;

  FILE *fd = fopen ( acct_fname, "r" ) ;

  if ( fd == NULL )
  {
    fl_alert ( "Cannot open the file for that account." ) ;
    valid_account = FALSE ;
    strcpy ( acct_name, " " ) ;
    balance = 0 ;
    updateGUI () ;
    return ;
  }

  int  dollars,  cents ;
  int adollars, acents ;
  char str[1024];

  fgets(str,1024,fd);
  if ( sscanf ( str, "Name=[%[^]]]", acct_name ) != 1 )
  { fclose(fd);fail(str,1);return;}

  fgets(str,1024,fd);
  if ( sscanf ( str, "NextCheque=%d", &next_cheque ) != 1 )
  { fclose(fd);fail(str,2);return;}

  fgets(str,1024,fd);
  if ( sscanf ( str, "Addrs1=[%[^]]]", addrs_line1 ) != 1 )
  { fclose(fd);fail(str,3);return;}

  fgets(str,1024,fd);
  if ( sscanf ( str, "Addrs2=[%[^]]]", addrs_line2 ) != 1 )
  { fclose(fd);fail(str,4);return;}

  fgets(str,1024,fd);
  if ( sscanf ( str, "NextAllowanceDue=%ld",&allowance_time ) != 1 )
  { fclose(fd);fail(str,5);return;}

  fgets(str,1024,fd);
  if ( sscanf ( str, "WeeklyInterestRate=%d%%", & interest_rate ) != 1 )
  { fclose(fd);fail(str,6);return;}

  fgets(str,1024,fd);
  if ( sscanf ( str, "Allowance=$%d.%d", &adollars,&acents)!= 2 )
  { fclose(fd);fail(str,7);return;}

  fgets(str,1024,fd);
  if ( sscanf ( str, "Balance=$%d.%d", & dollars,& cents)!= 2 )
  { fclose(fd);fail(str,8);return;}

  emptyStatement () ;

  while ( fgets(str,1024,fd) )
    addToStatement ( str ) ;

  fclose ( fd ) ;

  balance   =  dollars * 100 +  cents ;
  allowance = adollars * 100 + acents ;

  time_t now = time ( NULL ) ;

  if ( allowance_time <= 0 )
    allowance_time = now ;

  while ( now - allowance_time > ONE_WEEK )
  {
    int interest = balance * interest_rate / 100 ;

    balance += interest ;

    if ( interest > 0 )
    {
      sprintf ( str, "Interest Earned at %d%%", interest_rate ) ;
      addToStatement ( allowance_time, str, interest ) ;
    }
    else
    if ( interest < 0 )
    {
      sprintf ( str, "OVERDRAWN: Interest Charged at %d%%", interest_rate ) ;
      addToStatement ( allowance_time, str, interest ) ;
    }

    allowance_time += ONE_WEEK ;

    balance += allowance ;

    addToStatement ( allowance_time, "Weekly Allowance", allowance ) ;
  }

  valid_account = TRUE ;
  refreshAccount () ;
}



void createAccount ()
{
  time_t now = time ( NULL ) ;

  acct_fname [ 0 ] = '\0' ;
  strcpy ( acct_name  , "No Name Selected"      ) ;
  strcpy ( addrs_line1, "No Street Address"     ) ;
  strcpy ( addrs_line2, "No City/State/Zipcode" ) ;

  allowance_time = now ;
  allowance      = 1000 ;  /* $10 */
  balance        = 0 ;
  next_cheque    = 0 ;
  interest_rate  = 3 ;

  char *home = getenv ( "BOD_HOME" ) ;
  if ( home == NULL || home[0] == '\0' )
    home = "." ;

  char *tmp = fl_file_chooser ( "New Account Filename:",
                                "*.bod",
                                home ) ;

  if ( tmp == NULL || tmp[0] == '\0' )
  {
    valid_account = FALSE ;
    strcpy ( acct_name, " " ) ;
    balance = 0 ;
    updateGUI () ;
    return ;
  }

  emptyStatement () ;
  addToStatement ( now, "ACCOUNT OPENED", balance ) ;
  strcpy ( acct_fname, tmp ) ;
  valid_account = TRUE ;
  refreshAccount () ;
  editAccount () ;
}



void deleteAccount ()
{
  if ( ! check_account_name () ) return ;

  char message [ 1024 ] ;

  if ( balance != 0 )
  {
    if ( balance > 0 )
      sprintf ( message, "This account still has $%d.%02d left in it,\n"
                         "Are you sure you want to delete it?",
                           balance/100, balance-(balance/100)*100 ) ;
    else
      sprintf ( message, "This account owes the bank $%d.%02d,\n"
                         "Are you sure you want to delete it?",
                           (-balance)/100, (-balance)-((-balance)/100)*100 ) ;
  }
  else
    sprintf ( message, "Are you sure you want to delete this account?" ) ;

  if ( ! fl_ask ( message ) )
    return ;

  unlink ( acct_fname ) ;
  valid_account = FALSE ;
  strcpy ( acct_name, " " ) ;
  balance = 0 ;
  updateGUI () ;
}


void viewAllAccounts ()
{
}

void viewStatement ()
{
  if ( ! check_account_name () ) return ;

  if ( ! fl_ask ( "Print Statement?" ) )
    return ;

  printPostscriptStatement () ;
  fl_alert ( "Printing..." ) ;
}

void depositCash ()
{
  if ( ! check_account_name () ) return ;

  CashInterface *ch_gui = new CashInterface ( 0, 0, 100, 100 ) ;
  ch_gui   -> end  () ;
  ch_gui   -> Memo   -> value ( " " ) ;
  ch_gui   -> Amount -> value ( "$  0.00" ) ;
  ch_gui   -> mainwindow -> set_modal() ;
  ch_gui   -> show () ;
  exit_edit_mode = FALSE ;

  while ( ! exit_edit_mode )
  {
    if ( ! Fl::check () )
    {
      refreshAccount () ;
      delete ch_gui ;
      return ;
    }

    ch_gui->redraw () ;
  }

  char amountstr [ 1024 ] ;
  char *as = amountstr ;
  strcpy ( amountstr, ch_gui -> Amount -> value () ) ;

  while ( (isspace(*as) || *as == '$') && *as != '\0' ) as++ ;

  int amountnum = (int)( atof(as)*100.0f ) ;

  if ( amountnum > 0 )
  {
    time_t now = time ( NULL ) ;
    char memo [ 1024 ] ;

    sprintf ( memo, "CASH DEPOSIT - %s",
                             ch_gui -> Memo -> value () ) ;

    balance += amountnum ;
    addToStatement ( now, memo, amountnum ) ;
  }
  else
    fl_alert ( "Illegal input - cash NOT deposited" ) ;

  refreshAccount () ;
  delete ch_gui->mainwindow ;
  delete ch_gui ;
}

void withdrawCash ()
{
  if ( ! check_account_name () ) return ;

  CashInterface *ch_gui = new CashInterface ( 0, 0, 100, 100 ) ;
  ch_gui   -> end  () ;
  ch_gui   -> Memo   -> value ( " " ) ;
  ch_gui   -> Amount -> value ( "$  0.00" ) ;
  ch_gui   -> mainwindow -> set_modal() ;
  ch_gui   -> show () ;
  exit_edit_mode = FALSE ;

  while ( ! exit_edit_mode )
  {
    if ( ! Fl::check () )
    {
      refreshAccount () ;
      delete ch_gui ;
      return ;
    }

    ch_gui->redraw () ;
  }

  char amountstr [ 1024 ] ;
  char *as = amountstr ;
  strcpy ( amountstr, ch_gui -> Amount -> value () ) ;

  while ( (isspace(*as) || *as == '$') && *as != '\0' ) as++ ;

  int amountnum = (int)( atof(as)*100.0f ) ;

  if ( amountnum > 0 )
  {
    time_t now = time ( NULL ) ;
    char memo [ 1024 ] ;

    if ( balance < amountnum )
    {
      fl_alert ( "Insufficient funds to pay this cash." ) ;
      delete ch_gui->mainwindow ;
      delete ch_gui ;
      return ;
    }

    sprintf ( memo, "CASH WITHDRAWAL - %s",
                             ch_gui -> Memo -> value () ) ;

    balance -= amountnum ;
    addToStatement ( now, memo, -amountnum ) ;
  }
  else
    fl_alert ( "Illegal input - cash NOT withdrawn" ) ;

  refreshAccount () ;
  delete ch_gui->mainwindow ;
  delete ch_gui ;
}

void withdrawCheque () 
{
  if ( ! check_account_name () ) return ;

  ChequeInterface *ch_gui = new ChequeInterface ( 0, 0, 100, 100 ) ;
  ch_gui   -> end  () ;
  ch_gui   -> Number -> value ( "000" ) ;
  ch_gui   -> Memo   -> value ( " " ) ;
  ch_gui   -> Amount -> value ( "$  0.00" ) ;
  ch_gui   -> mainwindow -> set_modal() ;
  ch_gui   -> show () ;
  exit_edit_mode = FALSE ;

  while ( ! exit_edit_mode )
  {
    if ( ! Fl::check () )
    {
      refreshAccount () ;
      delete ch_gui ;
      return ;
    }

    ch_gui->redraw () ;
  }

  char amountstr [ 1024 ] ;
  char *as = amountstr ;
  strcpy ( amountstr, ch_gui -> Amount -> value () ) ;

  while ( (isspace(*as) || *as == '$') && *as != '\0' ) as++ ;

  int amountnum = (int)( atof(as)*100.0f ) ;
  int chequenum = atoi ( ch_gui -> Number -> value () ) ;

  if ( amountnum > 0 )
  {
    time_t now = time ( NULL ) ;
    char memo [ 1024 ] ;

    if ( balance < amountnum &&
         ! fl_ask ( "Insufficient funds to pay this cheque"
                    "Do you still want to pay it?" ) )
    {
      sprintf ( memo, "BOUNCED CHEQUE %03d! - %s", chequenum,
			   ch_gui -> Memo -> value () ) ;

      addToStatement ( now, memo, 0 ) ;
      refreshAccount () ;
      delete ch_gui->mainwindow ;
      delete ch_gui ;
      return ;
    }

    sprintf ( memo, "CHEQUE %03d - %s", chequenum,
                             ch_gui -> Memo -> value () ) ;

    balance -= amountnum ;
    addToStatement ( now, memo, -amountnum ) ;
  }
  else
    fl_alert ( "Illegal input - cheque NOT entered" ) ;

  refreshAccount () ;
  delete ch_gui->mainwindow ;
  delete ch_gui ;
}


